@@ -4,8 +4,8 @@ use log::debug;
 use termcolor::Color::{self, Cyan, Green, Red};
 
 use crate::core::registry::PackageRegistry;
-use crate::core::resolver::ResolveOpts;
-use crate::core::PackageId;
+use crate::core::resolver::features::{CliFeatures, HasDevUnits};
+use crate::core::{PackageId, PackageIdSpec};
 use crate::core::{Resolve, SourceId, Workspace};
 use crate::ops;
 use crate::util::config::Config;
@@ -17,20 +17,22 @@ pub struct UpdateOptions<'a> {
     pub precise: Option<&'a str>,
     pub aggressive: bool,
     pub dry_run: bool,
+    pub workspace: bool,
 }
 
 pub fn generate_lockfile(ws: &Workspace<'_>) -> CargoResult<()> {
     let mut registry = PackageRegistry::new(ws.config())?;
-    let resolve = ops::resolve_with_previous(
+    let mut resolve = ops::resolve_with_previous(
         &mut registry,
         ws,
-        ResolveOpts::everything(),
+        &CliFeatures::new_all(true),
+        HasDevUnits::Yes,
         None,
         None,
         &[],
         true,
     )?;
-    ops::write_pkg_lockfile(ws, &resolve)?;
+    ops::write_pkg_lockfile(ws, &mut resolve)?;
     Ok(())
 }
 
@@ -43,10 +45,6 @@ pub fn update_lockfile(ws: &Workspace<'_>, opts: &UpdateOptions<'_>) -> CargoRes
         anyhow::bail!("you can't generate a lockfile for an empty workspace.")
     }
 
-    if opts.config.offline() {
-        anyhow::bail!("you can't update in the offline mode");
-    }
-
     // Updates often require a lot of modifications to the registry, so ensure
     // that we're synchronized against other Cargos.
     let _lock = ws.config().acquire_package_cache_lock()?;
@@ -64,7 +62,8 @@ pub fn update_lockfile(ws: &Workspace<'_>, opts: &UpdateOptions<'_>) -> CargoRes
                     ops::resolve_with_previous(
                         &mut registry,
                         ws,
-                        ResolveOpts::everything(),
+                        &CliFeatures::new_all(true),
+                        HasDevUnits::Yes,
                         None,
                         None,
                         &[],
@@ -78,7 +77,10 @@ pub fn update_lockfile(ws: &Workspace<'_>, opts: &UpdateOptions<'_>) -> CargoRes
     let mut to_avoid = HashSet::new();
 
     if opts.to_update.is_empty() {
-        to_avoid.extend(previous_resolve.iter());
+        if !opts.workspace {
+            to_avoid.extend(previous_resolve.iter());
+            to_avoid.extend(previous_resolve.unused_patches());
+        }
     } else {
         let mut sources = Vec::new();
         for name in opts.to_update.iter() {
@@ -102,15 +104,21 @@ pub fn update_lockfile(ws: &Workspace<'_>, opts: &UpdateOptions<'_>) -> CargoRes
                     None => dep.source_id().with_precise(None),
                 });
             }
+            if let Ok(unused_id) =
+                PackageIdSpec::query_str(name, previous_resolve.unused_patches().iter().cloned())
+            {
+                to_avoid.insert(unused_id);
+            }
         }
 
         registry.add_sources(sources)?;
     }
 
-    let resolve = ops::resolve_with_previous(
+    let mut resolve = ops::resolve_with_previous(
         &mut registry,
         ws,
-        ResolveOpts::everything(),
+        &CliFeatures::new_all(true),
+        HasDevUnits::Yes,
         Some(&previous_resolve),
         Some(&to_avoid),
         &[],
@@ -147,7 +155,7 @@ pub fn update_lockfile(ws: &Workspace<'_>, opts: &UpdateOptions<'_>) -> CargoRes
             .shell()
             .warn("not updating lockfile due to dry run")?;
     } else {
-        ops::write_pkg_lockfile(ws, &resolve)?;
+        ops::write_pkg_lockfile(ws, &mut resolve)?;
     }
     return Ok(());
 
