@@ -1,30 +1,23 @@
+//! Utilities for handling git repositories, mainly around
+//! authentication/cloning.
+
 use crate::core::GitReference;
-use crate::util::errors::{CargoResult, CargoResultExt};
-use crate::util::paths;
-use crate::util::process_builder::process;
-use crate::util::{internal, network, Config, IntoUrl, Progress};
-use curl::easy::{Easy, List};
-use git2::{self, ObjectType};
+use crate::util::errors::CargoResult;
+use crate::util::{network, Config, IntoUrl, MetricsCounter, Progress};
+use anyhow::{anyhow, Context as _};
+use cargo_util::{paths, ProcessBuilder};
+use curl::easy::List;
+use git2::{self, ErrorClass, ObjectType};
 use log::{debug, info};
 use serde::ser;
 use serde::Serialize;
 use std::env;
 use std::fmt;
-use std::fs::File;
-use std::mem;
 use std::path::{Path, PathBuf};
 use std::process::Command;
+use std::time::{Duration, Instant};
 use url::Url;
 
-#[derive(PartialEq, Clone, Debug)]
-pub struct GitRevision(git2::Oid);
-
-impl ser::Serialize for GitRevision {
-    fn serialize<S: ser::Serializer>(&self, s: S) -> Result<S::Ok, S::Error> {
-        serialize_str(self, s)
-    }
-}
-
 fn serialize_str<T, S>(t: &T, s: S) -> Result<S::Ok, S::Error>
 where
     T: fmt::Display,
@@ -33,12 +26,6 @@ where
     s.collect_str(t)
 }
 
-impl fmt::Display for GitRevision {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        fmt::Display::fmt(&self.0, f)
-    }
-}
-
 pub struct GitShortID(git2::Buf);
 
 impl GitShortID {
@@ -67,12 +54,13 @@ pub struct GitDatabase {
 
 /// `GitCheckout` is a local checkout of a particular revision. Calling
 /// `clone_into` with a reference will resolve the reference into a revision,
-/// and return a `anyhow::Error` if no revision for that reference was found.
+/// and return an `anyhow::Error` if no revision for that reference was found.
 #[derive(Serialize)]
 pub struct GitCheckout<'a> {
     database: &'a GitDatabase,
     location: PathBuf,
-    revision: GitRevision,
+    #[serde(serialize_with = "serialize_str")]
+    revision: git2::Oid,
     #[serde(skip_serializing)]
     repo: git2::Repository,
 }
@@ -88,33 +76,52 @@ impl GitRemote {
         &self.url
     }
 
-    pub fn rev_for(&self, path: &Path, reference: &GitReference) -> CargoResult<GitRevision> {
+    pub fn rev_for(&self, path: &Path, reference: &GitReference) -> CargoResult<git2::Oid> {
         reference.resolve(&self.db_at(path)?.repo)
     }
 
     pub fn checkout(
         &self,
         into: &Path,
+        db: Option<GitDatabase>,
         reference: &GitReference,
+        locked_rev: Option<git2::Oid>,
         cargo_config: &Config,
-    ) -> CargoResult<(GitDatabase, GitRevision)> {
-        let mut repo_and_rev = None;
-        if let Ok(mut repo) = git2::Repository::open(into) {
-            self.fetch_into(&mut repo, cargo_config)
-                .chain_err(|| format!("failed to fetch into {}", into.display()))?;
-            if let Ok(rev) = reference.resolve(&repo) {
-                repo_and_rev = Some((repo, rev));
+    ) -> CargoResult<(GitDatabase, git2::Oid)> {
+        // If we have a previous instance of `GitDatabase` then fetch into that
+        // if we can. If that can successfully load our revision then we've
+        // populated the database with the latest version of `reference`, so
+        // return that database and the rev we resolve to.
+        if let Some(mut db) = db {
+            fetch(&mut db.repo, self.url.as_str(), reference, cargo_config)
+                .context(format!("failed to fetch into: {}", into.display()))?;
+            match locked_rev {
+                Some(rev) => {
+                    if db.contains(rev) {
+                        return Ok((db, rev));
+                    }
+                }
+                None => {
+                    if let Ok(rev) = reference.resolve(&db.repo) {
+                        return Ok((db, rev));
+                    }
+                }
             }
         }
-        let (repo, rev) = match repo_and_rev {
-            Some(pair) => pair,
-            None => {
-                let repo = self
-                    .clone_into(into, cargo_config)
-                    .chain_err(|| format!("failed to clone into: {}", into.display()))?;
-                let rev = reference.resolve(&repo)?;
-                (repo, rev)
-            }
+
+        // Otherwise start from scratch to handle corrupt git repositories.
+        // After our fetch (which is interpreted as a clone now) we do the same
+        // resolution to figure out what we cloned.
+        if into.exists() {
+            paths::remove_dir_all(into)?;
+        }
+        paths::create_dir_all(into)?;
+        let mut repo = init(into, true)?;
+        fetch(&mut repo, self.url.as_str(), reference, cargo_config)
+            .context(format!("failed to clone into: {}", into.display()))?;
+        let rev = match locked_rev {
+            Some(rev) => rev,
+            None => reference.resolve(&repo)?,
         };
 
         Ok((
@@ -135,49 +142,32 @@ impl GitRemote {
             repo,
         })
     }
-
-    fn fetch_into(&self, dst: &mut git2::Repository, cargo_config: &Config) -> CargoResult<()> {
-        // Create a local anonymous remote in the repository to fetch the url
-        let refspec = "refs/heads/*:refs/heads/*";
-        fetch(dst, self.url.as_str(), refspec, cargo_config)
-    }
-
-    fn clone_into(&self, dst: &Path, cargo_config: &Config) -> CargoResult<git2::Repository> {
-        if dst.exists() {
-            paths::remove_dir_all(dst)?;
-        }
-        paths::create_dir_all(dst)?;
-        let mut repo = init(dst, true)?;
-        fetch(
-            &mut repo,
-            self.url.as_str(),
-            "refs/heads/*:refs/heads/*",
-            cargo_config,
-        )?;
-        Ok(repo)
-    }
 }
 
 impl GitDatabase {
     pub fn copy_to(
         &self,
-        rev: GitRevision,
+        rev: git2::Oid,
         dest: &Path,
         cargo_config: &Config,
     ) -> CargoResult<GitCheckout<'_>> {
         let mut checkout = None;
         if let Ok(repo) = git2::Repository::open(dest) {
-            let mut co = GitCheckout::new(dest, self, rev.clone(), repo);
+            let mut co = GitCheckout::new(dest, self, rev, repo);
             if !co.is_fresh() {
-                // After a successful fetch operation do a sanity check to
-                // ensure we've got the object in our database to reset to. This
-                // can fail sometimes for corrupt repositories where the fetch
-                // operation succeeds but the object isn't actually there.
+                // After a successful fetch operation the subsequent reset can
+                // fail sometimes for corrupt repositories where the fetch
+                // operation succeeds but the object isn't actually there in one
+                // way or another. In these situations just skip the error and
+                // try blowing away the whole repository and trying with a
+                // clone.
                 co.fetch(cargo_config)?;
-                if co.has_object() {
-                    co.reset(cargo_config)?;
-                    assert!(co.is_fresh());
-                    checkout = Some(co);
+                match co.reset(cargo_config) {
+                    Ok(()) => {
+                        assert!(co.is_fresh());
+                        checkout = Some(co);
+                    }
+                    Err(e) => debug!("failed reset after fetch {:?}", e),
                 }
             } else {
                 checkout = Some(co);
@@ -191,37 +181,54 @@ impl GitDatabase {
         Ok(checkout)
     }
 
-    pub fn to_short_id(&self, revision: &GitRevision) -> CargoResult<GitShortID> {
-        let obj = self.repo.find_object(revision.0, None)?;
+    pub fn to_short_id(&self, revision: git2::Oid) -> CargoResult<GitShortID> {
+        let obj = self.repo.find_object(revision, None)?;
         Ok(GitShortID(obj.short_id()?))
     }
 
-    pub fn has_ref(&self, reference: &str) -> CargoResult<()> {
-        self.repo.revparse_single(reference)?;
-        Ok(())
+    pub fn contains(&self, oid: git2::Oid) -> bool {
+        self.repo.revparse_single(&oid.to_string()).is_ok()
+    }
+
+    pub fn resolve(&self, r: &GitReference) -> CargoResult<git2::Oid> {
+        r.resolve(&self.repo)
     }
 }
 
 impl GitReference {
-    fn resolve(&self, repo: &git2::Repository) -> CargoResult<GitRevision> {
-        let id = match *self {
-            GitReference::Tag(ref s) => (|| -> CargoResult<git2::Oid> {
-                let refname = format!("refs/tags/{}", s);
+    pub fn resolve(&self, repo: &git2::Repository) -> CargoResult<git2::Oid> {
+        let id = match self {
+            // Note that we resolve the named tag here in sync with where it's
+            // fetched into via `fetch` below.
+            GitReference::Tag(s) => (|| -> CargoResult<git2::Oid> {
+                let refname = format!("refs/remotes/origin/tags/{}", s);
                 let id = repo.refname_to_id(&refname)?;
                 let obj = repo.find_object(id, None)?;
                 let obj = obj.peel(ObjectType::Commit)?;
                 Ok(obj.id())
             })()
-            .chain_err(|| format!("failed to find tag `{}`", s))?,
-            GitReference::Branch(ref s) => {
+            .with_context(|| format!("failed to find tag `{}`", s))?,
+
+            // Resolve the remote name since that's all we're configuring in
+            // `fetch` below.
+            GitReference::Branch(s) => {
+                let name = format!("origin/{}", s);
                 let b = repo
-                    .find_branch(s, git2::BranchType::Local)
-                    .chain_err(|| format!("failed to find branch `{}`", s))?;
+                    .find_branch(&name, git2::BranchType::Remote)
+                    .with_context(|| format!("failed to find branch `{}`", s))?;
                 b.get()
                     .target()
                     .ok_or_else(|| anyhow::format_err!("branch `{}` did not have a target", s))?
             }
-            GitReference::Rev(ref s) => {
+
+            // We'll be using the HEAD commit
+            GitReference::DefaultBranch => {
+                let head_id = repo.refname_to_id("refs/remotes/origin/HEAD")?;
+                let head = repo.find_object(head_id, None)?;
+                head.peel(ObjectType::Commit)?.id()
+            }
+
+            GitReference::Rev(s) => {
                 let obj = repo.revparse_single(s)?;
                 match obj.as_tag() {
                     Some(tag) => tag.target_id(),
@@ -229,7 +236,7 @@ impl GitReference {
                 }
             }
         };
-        Ok(GitRevision(id))
+        Ok(id)
     }
 }
 
@@ -237,7 +244,7 @@ impl<'a> GitCheckout<'a> {
     fn new(
         path: &Path,
         database: &'a GitDatabase,
-        revision: GitRevision,
+        revision: git2::Oid,
         repo: git2::Repository,
     ) -> GitCheckout<'a> {
         GitCheckout {
@@ -251,7 +258,7 @@ impl<'a> GitCheckout<'a> {
     fn clone_into(
         into: &Path,
         database: &'a GitDatabase,
-        revision: GitRevision,
+        revision: git2::Oid,
         config: &Config,
     ) -> CargoResult<GitCheckout<'a>> {
         let dirname = into.parent().unwrap();
@@ -283,7 +290,6 @@ impl<'a> GitCheckout<'a> {
                 .clone_local(git2::build::CloneLocal::Local)
                 .with_checkout(checkout)
                 .fetch_options(fopts)
-                // .remote_create(|repo, _name, url| repo.remote_anonymous(url))
                 .clone(url.as_str(), into)?;
             repo = Some(r);
             Ok(())
@@ -297,7 +303,7 @@ impl<'a> GitCheckout<'a> {
 
     fn is_fresh(&self) -> bool {
         match self.repo.revparse_single("HEAD") {
-            Ok(ref head) if head.id() == self.revision.0 => {
+            Ok(ref head) if head.id() == self.revision => {
                 // See comments in reset() for why we check this
                 self.location.join(".cargo-ok").exists()
             }
@@ -308,15 +314,11 @@ impl<'a> GitCheckout<'a> {
     fn fetch(&mut self, cargo_config: &Config) -> CargoResult<()> {
         info!("fetch {}", self.repo.path().display());
         let url = self.database.path.into_url()?;
-        let refspec = "refs/heads/*:refs/heads/*";
-        fetch(&mut self.repo, url.as_str(), refspec, cargo_config)?;
+        let reference = GitReference::Rev(self.revision.to_string());
+        fetch(&mut self.repo, url.as_str(), &reference, cargo_config)?;
         Ok(())
     }
 
-    fn has_object(&self) -> bool {
-        self.repo.find_object(self.revision.0, None).is_ok()
-    }
-
     fn reset(&self, config: &Config) -> CargoResult<()> {
         // If we're interrupted while performing this reset (e.g., we die because
         // of a signal) Cargo needs to be sure to try to check out this repo
@@ -329,9 +331,15 @@ impl<'a> GitCheckout<'a> {
         let ok_file = self.location.join(".cargo-ok");
         let _ = paths::remove_file(&ok_file);
         info!("reset {} to {}", self.repo.path().display(), self.revision);
-        let object = self.repo.find_object(self.revision.0, None)?;
+
+        // Ensure libgit2 won't mess with newlines when we vendor.
+        if let Ok(mut git_config) = self.repo.config() {
+            git_config.set_bool("core.autocrlf", false)?;
+        }
+
+        let object = self.repo.find_object(self.revision, None)?;
         reset(&self.repo, &object, config)?;
-        File::create(ok_file)?;
+        paths::create(ok_file)?;
         Ok(())
     }
 
@@ -342,7 +350,7 @@ impl<'a> GitCheckout<'a> {
             info!("update submodules for: {:?}", repo.workdir().unwrap());
 
             for mut child in repo.submodules()? {
-                update_submodule(repo, &mut child, cargo_config).chain_err(|| {
+                update_submodule(repo, &mut child, cargo_config).with_context(|| {
                     format!(
                         "failed to update submodule `{}`",
                         child.name().unwrap_or("")
@@ -358,9 +366,9 @@ impl<'a> GitCheckout<'a> {
             cargo_config: &Config,
         ) -> CargoResult<()> {
             child.init(false)?;
-            let url = child
-                .url()
-                .ok_or_else(|| internal("non-utf8 url for submodule"))?;
+            let url = child.url().ok_or_else(|| {
+                anyhow::format_err!("non-utf8 url for submodule {:?}?", child.path())
+            })?;
 
             // A submodule which is listed in .gitmodules but not actually
             // checked out will not have a head id, so we should ignore it.
@@ -391,13 +399,16 @@ impl<'a> GitCheckout<'a> {
                 }
             };
             // Fetch data from origin and reset to the head commit
-            let refspec = "refs/heads/*:refs/heads/*";
-            fetch(&mut repo, url, refspec, cargo_config).chain_err(|| {
-                internal(format!(
+            let reference = GitReference::Rev(head.to_string());
+            cargo_config
+                .shell()
+                .status("Updating", format!("git submodule `{}`", url))?;
+            fetch(&mut repo, url, &reference, cargo_config).with_context(|| {
+                format!(
                     "failed to fetch submodule `{}` from {}",
                     child.name().unwrap_or(""),
                     url
-                ))
+                )
             })?;
 
             let obj = repo.find_object(head, None)?;
@@ -446,9 +457,14 @@ where
     let mut ssh_agent_attempts = Vec::new();
     let mut any_attempts = false;
     let mut tried_sshkey = false;
+    let mut url_attempt = None;
 
+    let orig_url = url;
     let mut res = f(&mut |url, username, allowed| {
         any_attempts = true;
+        if url != orig_url {
+            url_attempt = Some(url.to_string());
+        }
         // libgit2's "USERNAME" authentication actually means that it's just
         // asking us for a username to keep going. This is currently only really
         // used for SSH authentication and isn't really an authentication type.
@@ -533,8 +549,7 @@ where
     // call our callback, `f`, in a loop here.
     if ssh_username_requested {
         debug_assert!(res.is_err());
-        let mut attempts = Vec::new();
-        attempts.push("git".to_string());
+        let mut attempts = vec![String::from("git")];
         if let Ok(s) = env::var("USER").or_else(|_| env::var("USERNAME")) {
             attempts.push(s);
         }
@@ -580,18 +595,26 @@ where
             }
         }
     }
-
-    if res.is_ok() || !any_attempts {
-        return res.map_err(From::from);
-    }
+    let mut err = match res {
+        Ok(e) => return Ok(e),
+        Err(e) => e,
+    };
 
     // In the case of an authentication failure (where we tried something) then
     // we try to give a more helpful error message about precisely what we
     // tried.
-    let res = res.map_err(anyhow::Error::from).chain_err(|| {
+    if any_attempts {
         let mut msg = "failed to authenticate when downloading \
                        repository"
             .to_string();
+
+        if let Some(attempt) = &url_attempt {
+            if url != attempt {
+                msg.push_str(": ");
+                msg.push_str(attempt);
+            }
+        }
+        msg.push('\n');
         if !ssh_agent_attempts.is_empty() {
             let names = ssh_agent_attempts
                 .iter()
@@ -599,37 +622,67 @@ where
                 .collect::<Vec<_>>()
                 .join(", ");
             msg.push_str(&format!(
-                "\nattempted ssh-agent authentication, but \
-                 none of the usernames {} succeeded",
+                "\n* attempted ssh-agent authentication, but \
+                 no usernames succeeded: {}",
                 names
             ));
         }
         if let Some(failed_cred_helper) = cred_helper_bad {
             if failed_cred_helper {
                 msg.push_str(
-                    "\nattempted to find username/password via \
+                    "\n* attempted to find username/password via \
                      git's `credential.helper` support, but failed",
                 );
             } else {
                 msg.push_str(
-                    "\nattempted to find username/password via \
+                    "\n* attempted to find username/password via \
                      `credential.helper`, but maybe the found \
                      credentials were incorrect",
                 );
             }
         }
-        msg
-    })?;
-    Ok(res)
+        msg.push_str("\n\n");
+        msg.push_str("if the git CLI succeeds then `net.git-fetch-with-cli` may help here\n");
+        msg.push_str("https://doc.rust-lang.org/cargo/reference/config.html#netgit-fetch-with-cli");
+        err = err.context(msg);
+
+    // Otherwise if we didn't even get to the authentication phase them we may
+    // have failed to set up a connection, in these cases hint on the
+    // `net.git-fetch-with-cli` configuration option.
+    } else if let Some(e) = err.downcast_ref::<git2::Error>() {
+        match e.class() {
+            ErrorClass::Net
+            | ErrorClass::Ssl
+            | ErrorClass::Submodule
+            | ErrorClass::FetchHead
+            | ErrorClass::Ssh
+            | ErrorClass::Callback
+            | ErrorClass::Http => {
+                let mut msg = "network failure seems to have happened\n".to_string();
+                msg.push_str(
+                    "if a proxy or similar is necessary `net.git-fetch-with-cli` may help here\n",
+                );
+                msg.push_str(
+                    "https://doc.rust-lang.org/cargo/reference/config.html#netgit-fetch-with-cli",
+                );
+                err = err.context(msg);
+            }
+            _ => {}
+        }
+    }
+
+    Err(err)
 }
 
 fn reset(repo: &git2::Repository, obj: &git2::Object<'_>, config: &Config) -> CargoResult<()> {
     let mut pb = Progress::new("Checkout", config);
     let mut opts = git2::build::CheckoutBuilder::new();
     opts.progress(|_, cur, max| {
-        drop(pb.tick(cur, max));
+        drop(pb.tick(cur, max, ""));
     });
+    debug!("doing reset");
     repo.reset(obj, git2::ResetType::Hard, Some(&mut opts))?;
+    debug!("reset done");
     Ok(())
 }
 
@@ -642,20 +695,56 @@ pub fn with_fetch_options(
     let mut progress = Progress::new("Fetch", config);
     network::with_retry(config, || {
         with_authentication(url, git_config, |f| {
+            let mut last_update = Instant::now();
             let mut rcb = git2::RemoteCallbacks::new();
+            // We choose `N=10` here to make a `300ms * 10slots ~= 3000ms`
+            // sliding window for tracking the data transfer rate (in bytes/s).
+            let mut counter = MetricsCounter::<10>::new(0, last_update);
             rcb.credentials(f);
-
             rcb.transfer_progress(|stats| {
+                let indexed_deltas = stats.indexed_deltas();
+                let msg = if indexed_deltas > 0 {
+                    // Resolving deltas.
+                    format!(
+                        ", ({}/{}) resolving deltas",
+                        indexed_deltas,
+                        stats.total_deltas()
+                    )
+                } else {
+                    // Receiving objects.
+                    //
+                    // # Caveat
+                    //
+                    // Progress bar relies on git2 calling `transfer_progress`
+                    // to update its transfer rate, but we cannot guarantee a
+                    // periodic call of that callback. Thus if we don't receive
+                    // any data for, say, 10 seconds, the rate will get stuck
+                    // and never go down to 0B/s.
+                    // In the future, we need to find away to update the rate
+                    // even when the callback is not called.
+                    let now = Instant::now();
+                    // Scrape a `received_bytes` to the counter every 300ms.
+                    if now - last_update > Duration::from_millis(300) {
+                        counter.add(stats.received_bytes(), now);
+                        last_update = now;
+                    }
+                    fn format_bytes(bytes: f32) -> (&'static str, f32) {
+                        static UNITS: [&str; 5] = ["", "Ki", "Mi", "Gi", "Ti"];
+                        let i = (bytes.log2() / 10.0).min(4.0) as usize;
+                        (UNITS[i], bytes / 1024_f32.powi(i as i32))
+                    }
+                    let (unit, rate) = format_bytes(counter.rate());
+                    format!(", {:.2}{}B/s", rate, unit)
+                };
                 progress
-                    .tick(stats.indexed_objects(), stats.total_objects())
+                    .tick(stats.indexed_objects(), stats.total_objects(), &msg)
                     .is_ok()
             });
 
             // Create a local anonymous remote in the repository to fetch the
             // url
             let mut opts = git2::FetchOptions::new();
-            opts.remote_callbacks(rcb)
-                .download_tags(git2::AutotagOption::All);
+            opts.remote_callbacks(rcb);
             cb(opts)
         })?;
         Ok(())
@@ -665,7 +754,7 @@ pub fn with_fetch_options(
 pub fn fetch(
     repo: &mut git2::Repository,
     url: &str,
-    refspec: &str,
+    reference: &GitReference,
     config: &Config,
 ) -> CargoResult<()> {
     if config.frozen() {
@@ -680,19 +769,10 @@ pub fn fetch(
 
     // If we're fetching from GitHub, attempt GitHub's special fast path for
     // testing if we've already got an up-to-date copy of the repository
-
-    if let Ok(url) = Url::parse(url) {
-        if url.host_str() == Some("github.com") {
-            if let Ok(oid) = repo.refname_to_id("refs/remotes/origin/master") {
-                let mut handle = config.http()?.borrow_mut();
-                debug!("attempting GitHub fast path for {}", url);
-                if github_up_to_date(&mut handle, &url, &oid) {
-                    return Ok(());
-                } else {
-                    debug!("fast path failed, falling back to a git fetch");
-                }
-            }
-        }
+    match github_up_to_date(repo, url, reference, config) {
+        Ok(true) => return Ok(()),
+        Ok(false) => {}
+        Err(e) => debug!("failed to check github {:?}", e),
     }
 
     // We reuse repositories quite a lot, so before we go through and update the
@@ -701,6 +781,34 @@ pub fn fetch(
     // request we're about to issue.
     maybe_gc_repo(repo)?;
 
+    // Translate the reference desired here into an actual list of refspecs
+    // which need to get fetched. Additionally record if we're fetching tags.
+    let mut refspecs = Vec::new();
+    let mut tags = false;
+    match reference {
+        // For branches and tags we can fetch simply one reference and copy it
+        // locally, no need to fetch other branches/tags.
+        GitReference::Branch(b) => {
+            refspecs.push(format!("refs/heads/{0}:refs/remotes/origin/{0}", b));
+        }
+        GitReference::Tag(t) => {
+            refspecs.push(format!("refs/tags/{0}:refs/remotes/origin/tags/{0}", t));
+        }
+
+        GitReference::DefaultBranch => {
+            refspecs.push(String::from("HEAD:refs/remotes/origin/HEAD"));
+        }
+
+        // For `rev` dependencies we don't know what the rev will point to. To
+        // handle this situation we fetch all branches and tags, and then we
+        // pray it's somewhere in there.
+        GitReference::Rev(_) => {
+            refspecs.push(String::from("refs/heads/*:refs/remotes/origin/*"));
+            refspecs.push(String::from("HEAD:refs/remotes/origin/HEAD"));
+            tags = true;
+        }
+    }
+
     // Unfortunately `libgit2` is notably lacking in the realm of authentication
     // when compared to the `git` command line. As a result, allow an escape
     // hatch for users that would prefer to use `git`-the-CLI for fetching
@@ -708,12 +816,15 @@ pub fn fetch(
     // flavors of authentication possible while also still giving us all the
     // speed and portability of using `libgit2`.
     if let Some(true) = config.net_config()?.git_fetch_with_cli {
-        return fetch_with_cli(repo, url, refspec, config);
+        return fetch_with_cli(repo, url, &refspecs, tags, config);
     }
 
     debug!("doing a fetch for {}", url);
     let git_config = git2::Config::open_default()?;
     with_fetch_options(&git_config, url, config, &mut |mut opts| {
+        if tags {
+            opts.download_tags(git2::AutotagOption::All);
+        }
         // The `fetch` operation here may fail spuriously due to a corrupt
         // repository. It could also fail, however, for a whole slew of other
         // reasons (aka network related reasons). We want Cargo to automatically
@@ -726,17 +837,18 @@ pub fn fetch(
         // blown away the repository, then we want to return the error as-is.
         let mut repo_reinitialized = false;
         loop {
-            debug!("initiating fetch of {} from {}", refspec, url);
+            debug!("initiating fetch of {:?} from {}", refspecs, url);
             let res = repo
                 .remote_anonymous(url)?
-                .fetch(&[refspec], Some(&mut opts), None);
+                .fetch(&refspecs, Some(&mut opts), None);
             let err = match res {
                 Ok(()) => break,
                 Err(e) => e,
             };
             debug!("fetch failed: {}", err);
 
-            if !repo_reinitialized && err.class() == git2::ErrorClass::Reference {
+            if !repo_reinitialized && matches!(err.class(), ErrorClass::Reference | ErrorClass::Odb)
+            {
                 repo_reinitialized = true;
                 debug!(
                     "looks like this is a corrupt repository, reinitializing \
@@ -756,16 +868,19 @@ pub fn fetch(
 fn fetch_with_cli(
     repo: &mut git2::Repository,
     url: &str,
-    refspec: &str,
+    refspecs: &[String],
+    tags: bool,
     config: &Config,
 ) -> CargoResult<()> {
-    let mut cmd = process("git");
-    cmd.arg("fetch")
-        .arg("--tags") // fetch all tags
-        .arg("--force") // handle force pushes
+    let mut cmd = ProcessBuilder::new("git");
+    cmd.arg("fetch");
+    if tags {
+        cmd.arg("--tags");
+    }
+    cmd.arg("--force") // handle force pushes
         .arg("--update-head-ok") // see discussion in #2078
         .arg(url)
-        .arg(refspec)
+        .args(refspecs)
         // If cargo is run by git (for example, the `exec` command in `git
         // rebase`), the GIT_DIR is set by git and will point to the wrong
         // location (this takes precedence over the cwd). Make sure this is
@@ -835,7 +950,7 @@ fn maybe_gc_repo(repo: &mut git2::Repository) -> CargoResult<()> {
             );
             if out.status.success() {
                 let new = git2::Repository::open(repo.path())?;
-                mem::replace(repo, new);
+                *repo = new;
                 return Ok(());
             }
         }
@@ -885,9 +1000,9 @@ fn init(path: &Path, bare: bool) -> CargoResult<git2::Repository> {
 /// made.
 ///
 /// This function will attempt to hit that fast path and verify that the `oid`
-/// is actually the current `master` branch of the repository. If `true` is
-/// returned then no update needs to be performed, but if `false` is returned
-/// then the standard update logic still needs to happen.
+/// is actually the current branch of the repository. If `true` is returned then
+/// no update needs to be performed, but if `false` is returned then the
+/// standard update logic still needs to happen.
 ///
 /// [1]: https://developer.github.com/v3/repos/commits/#get-the-sha-1-of-a-commit-reference
 ///
@@ -895,37 +1010,59 @@ fn init(path: &Path, bare: bool) -> CargoResult<git2::Repository> {
 /// just a fast path. As a result all errors are ignored in this function and we
 /// just return a `bool`. Any real errors will be reported through the normal
 /// update path above.
-fn github_up_to_date(handle: &mut Easy, url: &Url, oid: &git2::Oid) -> bool {
-    macro_rules! r#try {
-        ($e:expr) => {
-            match $e {
-                Some(e) => e,
-                None => return false,
-            }
-        };
+fn github_up_to_date(
+    repo: &mut git2::Repository,
+    url: &str,
+    reference: &GitReference,
+    config: &Config,
+) -> CargoResult<bool> {
+    let url = Url::parse(url)?;
+    if url.host_str() != Some("github.com") {
+        return Ok(false);
     }
 
+    let github_branch_name = match reference {
+        GitReference::Branch(branch) => branch,
+        GitReference::Tag(tag) => tag,
+        GitReference::DefaultBranch => "HEAD",
+        GitReference::Rev(_) => {
+            debug!("can't use github fast path with `rev`");
+            return Ok(false);
+        }
+    };
+
     // This expects GitHub urls in the form `github.com/user/repo` and nothing
     // else
-    let mut pieces = r#try!(url.path_segments());
-    let username = r#try!(pieces.next());
-    let repo = r#try!(pieces.next());
+    let mut pieces = url
+        .path_segments()
+        .ok_or_else(|| anyhow!("no path segments on url"))?;
+    let username = pieces
+        .next()
+        .ok_or_else(|| anyhow!("couldn't find username"))?;
+    let repository = pieces
+        .next()
+        .ok_or_else(|| anyhow!("couldn't find repository name"))?;
     if pieces.next().is_some() {
-        return false;
+        anyhow::bail!("too many segments on URL");
     }
 
-    let url = format!(
-        "https://api.github.com/repos/{}/{}/commits/master",
-        username, repo
-    );
-    r#try!(handle.get(true).ok());
-    r#try!(handle.url(&url).ok());
-    r#try!(handle.useragent("cargo").ok());
-    let mut headers = List::new();
-    r#try!(headers.append("Accept: application/vnd.github.3.sha").ok());
-    r#try!(headers.append(&format!("If-None-Match: \"{}\"", oid)).ok());
-    r#try!(handle.http_headers(headers).ok());
-    r#try!(handle.perform().ok());
+    // Trim off the `.git` from the repository, if present, since that's
+    // optional for GitHub and won't work when we try to use the API as well.
+    let repository = repository.strip_suffix(".git").unwrap_or(repository);
 
-    r#try!(handle.response_code().ok()) == 304
+    let url = format!(
+        "https://api.github.com/repos/{}/{}/commits/{}",
+        username, repository, github_branch_name,
+    );
+    let mut handle = config.http()?.borrow_mut();
+    debug!("attempting GitHub fast path for {}", url);
+    handle.get(true)?;
+    handle.url(&url)?;
+    handle.useragent("cargo")?;
+    let mut headers = List::new();
+    headers.append("Accept: application/vnd.github.3.sha")?;
+    headers.append(&format!("If-None-Match: \"{}\"", reference.resolve(repo)?))?;
+    handle.http_headers(headers)?;
+    handle.perform()?;
+    Ok(handle.response_code()? == 304)
 }
