@@ -1,11 +1,9 @@
 //! Tests for the `cargo new` command.
 
+use cargo_test_support::cargo_process;
+use cargo_test_support::paths;
 use std::env;
 use std::fs::{self, File};
-use std::io::prelude::*;
-
-use cargo_test_support::paths;
-use cargo_test_support::{cargo_process, git_process};
 
 fn create_empty_gitconfig() {
     // This helps on Windows where libgit2 is very aggressive in attempting to
@@ -17,7 +15,6 @@ fn create_empty_gitconfig() {
 #[cargo_test]
 fn simple_lib() {
     cargo_process("new --lib foo --vcs none --edition 2015")
-        .env("USER", "foo")
         .with_stderr("[CREATED] library `foo` package")
         .run();
 
@@ -27,18 +24,15 @@ fn simple_lib() {
     assert!(!paths::root().join("foo/.gitignore").is_file());
 
     let lib = paths::root().join("foo/src/lib.rs");
-    let mut contents = String::new();
-    File::open(&lib)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
+    let contents = fs::read_to_string(&lib).unwrap();
     assert_eq!(
         contents,
         r#"#[cfg(test)]
 mod tests {
     #[test]
     fn it_works() {
-        assert_eq!(2 + 2, 4);
+        let result = 2 + 2;
+        assert_eq!(result, 4);
     }
 }
 "#
@@ -50,7 +44,6 @@ mod tests {
 #[cargo_test]
 fn simple_bin() {
     cargo_process("new --bin foo --edition 2015")
-        .env("USER", "foo")
         .with_stderr("[CREATED] binary (application) `foo` package")
         .run();
 
@@ -67,7 +60,6 @@ fn simple_bin() {
 #[cargo_test]
 fn both_lib_and_bin() {
     cargo_process("new --lib --bin foo")
-        .env("USER", "foo")
         .with_status(101)
         .with_stderr("[ERROR] can't specify both lib and binary outputs")
         .run();
@@ -75,9 +67,7 @@ fn both_lib_and_bin() {
 
 #[cargo_test]
 fn simple_git() {
-    cargo_process("new --lib foo --edition 2015")
-        .env("USER", "foo")
-        .run();
+    cargo_process("new --lib foo --edition 2015").run();
 
     assert!(paths::root().is_dir());
     assert!(paths::root().join("foo/Cargo.toml").is_file());
@@ -86,11 +76,7 @@ fn simple_git() {
     assert!(paths::root().join("foo/.gitignore").is_file());
 
     let fp = paths::root().join("foo/.gitignore");
-    let mut contents = String::new();
-    File::open(&fp)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
+    let contents = fs::read_to_string(&fp).unwrap();
     assert_eq!(contents, "/target\nCargo.lock\n",);
 
     cargo_process("build").cwd(&paths::root().join("foo")).run();
@@ -128,8 +114,18 @@ fn invalid_characters() {
         .with_status(101)
         .with_stderr(
             "\
-[ERROR] Invalid character `.` in crate name: `foo.rs`
-use --name to override crate name",
+[ERROR] invalid character `.` in package name: `foo.rs`, [..]
+If you need a package name to not match the directory name, consider using --name flag.
+If you need a binary with the name \"foo.rs\", use a valid package name, \
+and set the binary name to be different from the package. \
+This can be done by setting the binary filename to `src/bin/foo.rs.rs` \
+or change the name in Cargo.toml with:
+
+    [bin]
+    name = \"foo.rs\"
+    path = \"src/main.rs\"
+
+",
         )
         .run();
 }
@@ -139,8 +135,19 @@ fn reserved_name() {
     cargo_process("new test")
         .with_status(101)
         .with_stderr(
-            "[ERROR] The name `test` cannot be used as a crate name\n\
-             use --name to override crate name",
+            "\
+[ERROR] the name `test` cannot be used as a package name, it conflicts [..]
+If you need a package name to not match the directory name, consider using --name flag.
+If you need a binary with the name \"test\", use a valid package name, \
+and set the binary name to be different from the package. \
+This can be done by setting the binary filename to `src/bin/test.rs` \
+or change the name in Cargo.toml with:
+
+    [bin]
+    name = \"test\"
+    path = \"src/main.rs\"
+
+",
         )
         .run();
 }
@@ -150,8 +157,20 @@ fn reserved_binary_name() {
     cargo_process("new --bin incremental")
         .with_status(101)
         .with_stderr(
-            "[ERROR] The name `incremental` cannot be used as a crate name\n\
-             use --name to override crate name",
+            "\
+[ERROR] the name `incremental` cannot be used as a package name, it conflicts [..]
+If you need a package name to not match the directory name, consider using --name flag.
+",
+        )
+        .run();
+
+    cargo_process("new --lib incremental")
+        .with_stderr(
+            "\
+[WARNING] the name `incremental` will not support binary executables with that name, \
+it conflicts with cargo's build directory names
+[CREATED] library `incremental` package
+",
         )
         .run();
 }
@@ -161,263 +180,78 @@ fn keyword_name() {
     cargo_process("new pub")
         .with_status(101)
         .with_stderr(
-            "[ERROR] The name `pub` cannot be used as a crate name\n\
-             use --name to override crate name",
+            "\
+[ERROR] the name `pub` cannot be used as a package name, it is a Rust keyword
+If you need a package name to not match the directory name, consider using --name flag.
+If you need a binary with the name \"pub\", use a valid package name, \
+and set the binary name to be different from the package. \
+This can be done by setting the binary filename to `src/bin/pub.rs` \
+or change the name in Cargo.toml with:
+
+    [bin]
+    name = \"pub\"
+    path = \"src/main.rs\"
+
+",
         )
         .run();
 }
 
 #[cargo_test]
-fn finds_author_user() {
-    create_empty_gitconfig();
-    cargo_process("new foo").env("USER", "foo").run();
+fn std_name() {
+    cargo_process("new core")
+        .with_stderr(
+            "\
+[WARNING] the name `core` is part of Rust's standard library
+It is recommended to use a different name to avoid problems.
+If you need a package name to not match the directory name, consider using --name flag.
+If you need a binary with the name \"core\", use a valid package name, \
+and set the binary name to be different from the package. \
+This can be done by setting the binary filename to `src/bin/core.rs` \
+or change the name in Cargo.toml with:
 
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["foo"]"#));
-}
+    [bin]
+    name = \"core\"
+    path = \"src/main.rs\"
 
-#[cargo_test]
-fn finds_author_user_escaped() {
-    create_empty_gitconfig();
-    cargo_process("new foo").env("USER", "foo \"bar\"").run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["foo \"bar\""]"#));
-}
-
-#[cargo_test]
-fn finds_author_username() {
-    create_empty_gitconfig();
-    cargo_process("new foo")
-        .env_remove("USER")
-        .env("USERNAME", "foo")
-        .run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["foo"]"#));
-}
-
-#[cargo_test]
-fn finds_author_name() {
-    create_empty_gitconfig();
-    cargo_process("new foo")
-        .env_remove("USERNAME")
-        .env("NAME", "foo")
-        .run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["foo"]"#));
-}
-
-#[cargo_test]
-fn finds_author_priority() {
-    cargo_process("new foo")
-        .env("USER", "bar2")
-        .env("EMAIL", "baz2")
-        .env("CARGO_NAME", "bar")
-        .env("CARGO_EMAIL", "baz")
-        .run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["bar <baz>"]"#));
-}
-
-#[cargo_test]
-fn finds_author_email() {
-    create_empty_gitconfig();
-    cargo_process("new foo")
-        .env("USER", "bar")
-        .env("EMAIL", "baz")
-        .run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["bar <baz>"]"#));
-}
-
-#[cargo_test]
-fn finds_author_git() {
-    git_process("config --global user.name bar").exec().unwrap();
-    git_process("config --global user.email baz")
-        .exec()
-        .unwrap();
-    cargo_process("new foo").env("USER", "foo").run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["bar <baz>"]"#));
-}
-
-#[cargo_test]
-fn finds_local_author_git() {
-    git_process("init").exec_with_output().unwrap();
-    git_process("config --global user.name foo").exec().unwrap();
-    git_process("config --global user.email foo@bar")
-        .exec()
-        .unwrap();
-
-    // Set local git user config
-    git_process("config user.name bar").exec().unwrap();
-    git_process("config user.email baz").exec().unwrap();
-    cargo_process("init").env("USER", "foo").run();
-
-    let toml = paths::root().join("Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["bar <baz>"]"#));
-}
-
-#[cargo_test]
-fn finds_git_author() {
-    cargo_process("new foo")
-        .env("GIT_AUTHOR_NAME", "foo")
-        .env("GIT_AUTHOR_EMAIL", "gitfoo")
-        .run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["foo <gitfoo>"]"#), contents);
-}
-
-#[cargo_test]
-fn finds_git_committer() {
-    create_empty_gitconfig();
-    cargo_process("new foo")
-        .env_remove("USER")
-        .env("GIT_COMMITTER_NAME", "foo")
-        .env("GIT_COMMITTER_EMAIL", "gitfoo")
-        .run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["foo <gitfoo>"]"#));
-}
-
-#[cargo_test]
-fn author_prefers_cargo() {
-    git_process("config --global user.name foo").exec().unwrap();
-    git_process("config --global user.email bar")
-        .exec()
-        .unwrap();
-    let root = paths::root();
-    fs::create_dir(&root.join(".cargo")).unwrap();
-    File::create(&root.join(".cargo/config"))
-        .unwrap()
-        .write_all(
-            br#"
-        [cargo-new]
-        name = "new-foo"
-        email = "new-bar"
-        vcs = "none"
-    "#,
+[CREATED] binary (application) `core` package
+",
         )
-        .unwrap();
-
-    cargo_process("new foo").env("USER", "foo").run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["new-foo <new-bar>"]"#));
-    assert!(!root.join("foo/.gitignore").exists());
-}
-
-#[cargo_test]
-fn strip_angle_bracket_author_email() {
-    create_empty_gitconfig();
-    cargo_process("new foo")
-        .env("USER", "bar")
-        .env("EMAIL", "<baz>")
         .run();
-
-    let toml = paths::root().join("foo/Cargo.toml");
-    let mut contents = String::new();
-    File::open(&toml)
-        .unwrap()
-        .read_to_string(&mut contents)
-        .unwrap();
-    assert!(contents.contains(r#"authors = ["bar <baz>"]"#));
 }
 
 #[cargo_test]
 fn git_prefers_command_line() {
     let root = paths::root();
     fs::create_dir(&root.join(".cargo")).unwrap();
-    File::create(&root.join(".cargo/config"))
-        .unwrap()
-        .write_all(
-            br#"
-        [cargo-new]
-        vcs = "none"
-        name = "foo"
-        email = "bar"
-    "#,
-        )
-        .unwrap();
+    fs::write(
+        &root.join(".cargo/config"),
+        r#"
+            [cargo-new]
+            vcs = "none"
+            name = "foo"
+            email = "bar"
+        "#,
+    )
+    .unwrap();
 
-    cargo_process("new foo --vcs git").env("USER", "foo").run();
+    cargo_process("new foo --vcs git").run();
     assert!(paths::root().join("foo/.gitignore").exists());
+    assert!(!fs::read_to_string(paths::root().join("foo/Cargo.toml"))
+        .unwrap()
+        .contains("authors ="));
 }
 
 #[cargo_test]
 fn subpackage_no_git() {
-    cargo_process("new foo").env("USER", "foo").run();
+    cargo_process("new foo").run();
 
     assert!(paths::root().join("foo/.git").is_dir());
     assert!(paths::root().join("foo/.gitignore").is_file());
 
     let subpackage = paths::root().join("foo").join("components");
     fs::create_dir(&subpackage).unwrap();
-    cargo_process("new foo/components/subcomponent")
-        .env("USER", "foo")
-        .run();
+    cargo_process("new foo/components/subcomponent").run();
 
     assert!(!paths::root()
         .join("foo/components/subcomponent/.git")
@@ -429,7 +263,7 @@ fn subpackage_no_git() {
 
 #[cargo_test]
 fn subpackage_git_with_gitignore() {
-    cargo_process("new foo").env("USER", "foo").run();
+    cargo_process("new foo").run();
 
     assert!(paths::root().join("foo/.git").is_dir());
     assert!(paths::root().join("foo/.gitignore").is_file());
@@ -439,9 +273,7 @@ fn subpackage_git_with_gitignore() {
 
     let subpackage = paths::root().join("foo/components");
     fs::create_dir(&subpackage).unwrap();
-    cargo_process("new foo/components/subcomponent")
-        .env("USER", "foo")
-        .run();
+    cargo_process("new foo/components/subcomponent").run();
 
     assert!(paths::root()
         .join("foo/components/subcomponent/.git")
@@ -453,13 +285,11 @@ fn subpackage_git_with_gitignore() {
 
 #[cargo_test]
 fn subpackage_git_with_vcs_arg() {
-    cargo_process("new foo").env("USER", "foo").run();
+    cargo_process("new foo").run();
 
     let subpackage = paths::root().join("foo").join("components");
     fs::create_dir(&subpackage).unwrap();
-    cargo_process("new foo/components/subcomponent --vcs git")
-        .env("USER", "foo")
-        .run();
+    cargo_process("new foo/components/subcomponent --vcs git").run();
 
     assert!(paths::root()
         .join("foo/components/subcomponent/.git")
@@ -483,66 +313,63 @@ fn unknown_flags() {
 fn explicit_invalid_name_not_suggested() {
     cargo_process("new --name 10-invalid a")
         .with_status(101)
-        .with_stderr("[ERROR] Package names starting with a digit cannot be used as a crate name")
+        .with_stderr(
+            "\
+[ERROR] the name `10-invalid` cannot be used as a package name, \
+the name cannot start with a digit\n\
+If you need a binary with the name \"10-invalid\", use a valid package name, \
+and set the binary name to be different from the package. \
+This can be done by setting the binary filename to `src/bin/10-invalid.rs` \
+or change the name in Cargo.toml with:
+
+    [bin]
+    name = \"10-invalid\"
+    path = \"src/main.rs\"
+
+",
+        )
         .run();
 }
 
 #[cargo_test]
 fn explicit_project_name() {
     cargo_process("new --lib foo --name bar")
-        .env("USER", "foo")
         .with_stderr("[CREATED] library `bar` package")
         .run();
 }
 
 #[cargo_test]
 fn new_with_edition_2015() {
-    cargo_process("new --edition 2015 foo")
-        .env("USER", "foo")
-        .run();
+    cargo_process("new --edition 2015 foo").run();
     let manifest = fs::read_to_string(paths::root().join("foo/Cargo.toml")).unwrap();
     assert!(manifest.contains("edition = \"2015\""));
 }
 
 #[cargo_test]
 fn new_with_edition_2018() {
-    cargo_process("new --edition 2018 foo")
-        .env("USER", "foo")
-        .run();
+    cargo_process("new --edition 2018 foo").run();
     let manifest = fs::read_to_string(paths::root().join("foo/Cargo.toml")).unwrap();
     assert!(manifest.contains("edition = \"2018\""));
 }
 
 #[cargo_test]
 fn new_default_edition() {
-    cargo_process("new foo").env("USER", "foo").run();
+    cargo_process("new foo").run();
     let manifest = fs::read_to_string(paths::root().join("foo/Cargo.toml")).unwrap();
-    assert!(manifest.contains("edition = \"2018\""));
+    assert!(manifest.contains("edition = \"2021\""));
 }
 
 #[cargo_test]
 fn new_with_bad_edition() {
     cargo_process("new --edition something_else foo")
-        .env("USER", "foo")
         .with_stderr_contains("error: 'something_else' isn't a valid value[..]")
         .with_status(1)
         .run();
 }
 
-#[cargo_test]
-fn new_with_blank_email() {
-    cargo_process("new foo")
-        .env("CARGO_NAME", "Sen")
-        .env("CARGO_EMAIL", "")
-        .run();
-
-    let contents = fs::read_to_string(paths::root().join("foo/Cargo.toml")).unwrap();
-    assert!(contents.contains(r#"authors = ["Sen"]"#), contents);
-}
-
 #[cargo_test]
 fn new_with_reference_link() {
-    cargo_process("new foo").env("USER", "foo").run();
+    cargo_process("new foo").run();
 
     let contents = fs::read_to_string(paths::root().join("foo/Cargo.toml")).unwrap();
     assert!(contents.contains("# See more keys and their definitions at https://doc.rust-lang.org/cargo/reference/manifest.html"))
@@ -550,7 +377,11 @@ fn new_with_reference_link() {
 
 #[cargo_test]
 fn lockfile_constant_during_new() {
-    cargo_process("new foo").env("USER", "foo").run();
+    if !cargo_test_support::is_nightly() {
+        // Remove when 1.56 is stable (cargo new defaults to 2021).
+        return;
+    }
+    cargo_process("new foo").run();
 
     cargo_process("build").cwd(&paths::root().join("foo")).run();
     let before = fs::read_to_string(paths::root().join("foo/Cargo.lock")).unwrap();
@@ -558,3 +389,108 @@ fn lockfile_constant_during_new() {
     let after = fs::read_to_string(paths::root().join("foo/Cargo.lock")).unwrap();
     assert_eq!(before, after);
 }
+
+#[cargo_test]
+fn restricted_windows_name() {
+    if cfg!(windows) {
+        cargo_process("new nul")
+            .with_status(101)
+            .with_stderr(
+                "\
+[ERROR] cannot use name `nul`, it is a reserved Windows filename
+If you need a package name to not match the directory name, consider using --name flag.
+",
+            )
+            .run();
+    } else {
+        cargo_process("new nul")
+            .with_stderr(
+                "\
+[WARNING] the name `nul` is a reserved Windows filename
+This package will not work on Windows platforms.
+[CREATED] binary (application) `nul` package
+",
+            )
+            .run();
+    }
+}
+
+#[cargo_test]
+fn non_ascii_name() {
+    cargo_process("new Привет")
+        .with_stderr(
+            "\
+[WARNING] the name `Привет` contains non-ASCII characters
+Support for non-ASCII crate names is experimental and only valid on the nightly toolchain.
+[CREATED] binary (application) `Привет` package
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn non_ascii_name_invalid() {
+    // These are alphanumeric characters, but not Unicode XID.
+    cargo_process("new ⒶⒷⒸ")
+        .with_status(101)
+        .with_stderr(
+            "\
+[ERROR] invalid character `Ⓐ` in package name: `ⒶⒷⒸ`, \
+the first character must be a Unicode XID start character (most letters or `_`)
+If you need a package name to not match the directory name, consider using --name flag.
+If you need a binary with the name \"ⒶⒷⒸ\", use a valid package name, \
+and set the binary name to be different from the package. \
+This can be done by setting the binary filename to `src/bin/ⒶⒷⒸ.rs` \
+or change the name in Cargo.toml with:
+
+    [bin]
+    name = \"ⒶⒷⒸ\"
+    path = \"src/main.rs\"
+
+",
+        )
+        .run();
+
+    cargo_process("new a¼")
+        .with_status(101)
+        .with_stderr(
+            "\
+[ERROR] invalid character `¼` in package name: `a¼`, \
+characters must be Unicode XID characters (numbers, `-`, `_`, or most letters)
+If you need a package name to not match the directory name, consider using --name flag.
+If you need a binary with the name \"a¼\", use a valid package name, \
+and set the binary name to be different from the package. \
+This can be done by setting the binary filename to `src/bin/a¼.rs` \
+or change the name in Cargo.toml with:
+
+    [bin]
+    name = \"a¼\"
+    path = \"src/main.rs\"
+
+",
+        )
+        .run();
+}
+
+#[cargo_test]
+fn git_default_branch() {
+    // Check for init.defaultBranch support.
+    create_empty_gitconfig();
+    cargo_process("new foo").run();
+    let repo = git2::Repository::open(paths::root().join("foo")).unwrap();
+    let head = repo.find_reference("HEAD").unwrap();
+    assert_eq!(head.symbolic_target().unwrap(), "refs/heads/master");
+
+    fs::write(
+        paths::home().join(".gitconfig"),
+        r#"
+        [init]
+            defaultBranch = hello
+        "#,
+    )
+    .unwrap();
+    cargo_process("new bar").run();
+    let repo = git2::Repository::open(paths::root().join("bar")).unwrap();
+    let head = repo.find_reference("HEAD").unwrap();
+    assert_eq!(head.symbolic_target().unwrap(), "refs/heads/hello");
+}
